unit KalmanFiltering;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Math;

type
  TMode = (mdRecursive, mdMovingaverage);
  TMatrixDbl = array of array of  Extended;
  TNotifyEventNextEstimation = procedure (Sender: TObject; Estimation: double) of object;
  TKalmanFiltering = class(TComponent)
  private
    { Private declarations }
    //  
    FFilterExponent: Integer;
    //   
    FMode: TMode;
    //   
    FSizeTeachingSequence: integer;
    //  -    
    FOnNextItem: TNotifyEventNextEstimation;
    //  -    
    FOnNextPrioriEstimation: TNotifyEventNextEstimation;
    //  -    
    FOnNextPosterioriEstimation: TNotifyEventNextEstimation;
    //  
    TransformMatrix : TMatrixDbl;
    // Work matrix H
    WorkHMatrix: TMatrixDbl;
    // Work matrix P
    WorkPMatrix: TMatrixDbl;
    // Inverse work matrix P
    InverseWorkPMatrix: TMatrixDbl;
    // Work matrix Hnull
    WorkHnullMatrix: TMatrixDbl;
    // Work matrix N
    WorkNMatrix: TMatrixDbl;
    TransposeWorkNMatrix: TMatrixDbl;
    //  TransientStateItems matrix
    TransientStateItems: TMatrixDbl;
    //  
    PrioriEstimation : TMatrixDbl;
    //  
    PosterioriEstimation: TMatrixDbl;
    //     n
    PresentObservableValue : Extended;
    //   
    ItemCounter : integer;
    //  -   
    procedure SetFilterExponent(Value: Integer);
    //  -   
    procedure SetSizeTeachingSequence(Value: Integer);
    // Initialization filter
    procedure InitializationFilter;
    // Calculation transient state
    procedure CalcTransientState;
    //   
    function FormingTransformMatrix(var TransformMatrix: TMatrixDbl; FilterExponent: Integer; TimeInterval: Integer): Integer;
    // Forming Work Matrix H
    function FormingWorkHMatrix(var WorkHMatrix: TMatrixDbl; FilterExponent: Integer): Integer;
    // Forming Work Matrix Hnull
    function FormingWorkHnullMatrix(var WorkHnullMatrix: TMatrixDbl; FilterExponent: Integer): Integer;
    // Forming Work Matriz P
    function FormingWorkPMatrix(WorkHMatrix: TMatrixDbl; var WorkPMatrix: TMatrixDbl): Integer;
    // Forming Work Matrix N
    function FormingWorkNMatrix(var WorkNMatrix: TMatrixDbl; ValueN: integer): integer;
    // Calculate priori estimation
    function CalcPrioriEstimation: Integer;
    // Get priori estimation  item
    function GetPrioriEstimationItem(Index: integer): double;
    // Caculate posteriori estimation
    function CalcPosterioriEstimation: Integer;
    // Get posteriori estimation item
    function GetPosterioriEstimationItem(Index: integer): double;
    //     LU - 
    function InverseSymmetryMatrix(InputMatrix: TMatrixDbl; var OutputMatrix: TMatrixDbl):Integer;
    //  
    function TransposeMatrix(InputMatrix: TMatrixDbl; var OutputMatrix : TMatrixDbl): Integer;
    //  
    function MultiplicationMatrix(MultiplicandMatrix, MultiplierMatrix: TMatrixDbl; var ProductMatrix: TMatrixDbl): Integer;
    // Sum Matrix's
    function SumMatrix(Item1Matrix: TMatrixDbl; Item2Matrix: TMatrixDbl; var SumMatrix: TMatrixDbl): Integer;
    // Copy Matrix
    function CopyMatrix(SourceMatrix: TMatrixDbl; var DestinationMatrix: TMatrixDbl): Integer;
  protected
    { Protected declarations }
  public
    { Public declarations }
    // 
    constructor Create(AOwner: TComponent); override;
    // 
    destructor Destroy; override;
    // 
   // procedure TestMatrix();
    //  
    procedure ClearFilter;
    //     
    procedure AddInputDataItem(Value: double);
    // To procced along the polynomial
    procedure GoAlongPolynomial;
    // Get priori estimation  item
    property PrioriEstimationItem[Index: integer]: double read GetPrioriEstimationItem;
    // Get posteriori estimation item
    property PosterioriEstimationItem[Index: integer]: double read GetPosterioriEstimationItem;

  published
    { Published declarations }
    //  -  
    property FilterExponent: integer read FFilterExponent write SetFilterExponent;
    //  " "
    property Mode: TMode read FMode write FMode;
    //  -   
    property SizeTeachingSequence: integer read FSizeTeachingSequence write SetSizeTeachingSequence;
    //  -    
    property OnNextItem: TNotifyEventNextEstimation read FOnNextItem write FOnNextItem;
    //  "   "
    property OnNextPrioriEstimation: TNotifyEventNextEstimation read FOnNextPrioriEstimation write FOnNextPrioriEstimation;
    //  "   "
    property OnNextPosterioriEstimation: TNotifyEventNextEstimation read FOnNextPosterioriEstimation write FOnNextPosterioriEstimation;
  end;

implementation

//////////////////
// constructor  //
//////////////////
constructor TKalmanFiltering.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFilterExponent := 2;
  FMode := mdRecursive;
end;

/////////////////
// destructor  //
/////////////////
destructor TKalmanFiltering.Destroy;
begin
  TransformMatrix := nil;
  PrioriEstimation := nil;
  PosterioriEstimation := nil;
  inherited Destroy;
end;

//////////////////////
//    //
//////////////////////
procedure TKalmanFiltering.ClearFilter();
begin
  ItemCounter := 0;
end;

///////////////////////////////////////
//  -    //
///////////////////////////////////////
procedure TKalmanFiltering.SetFilterExponent(Value: Integer);
begin
  if Value < 0 then
    FFilterExponent := 0
  else if Value > 10 then
         FFilterExponent := 10
       else
         FFilterExponent := value;
end;

///////////////////////////////////////////////
//  -     //
///////////////////////////////////////////////
procedure TKalmanFiltering.SetSizeTeachingSequence(Value: Integer);
begin
  if Value > 0 then
    FSizeTeachingSequence := Value
  else
    FSizeTeachingSequence := 0;
end;

///////////////////////////
// Initialization filter //
///////////////////////////
procedure TKalmanFiltering.InitializationFilter;
var i, j: Integer;
begin
  PrioriEstimation := nil;
  PosterioriEstimation := nil;
  TransientStateItems := nil;
  SetLength(PrioriEstimation, FFilterExponent + 1, 1);
  SetLength(PosterioriEstimation, FFilterExponent + 1, 1);
  SetLength(TransientStateItems, FFilterExponent + 1, 1);
  For i := 0 to High(PrioriEstimation) do
    For j:= 0 to High(PrioriEstimation[0]) do
      begin
        PrioriEstimation[i,j] := 0;
        PosterioriEstimation[i, j] := 0;
        TransientStateItems[i, j] := 0;
      end;
  FormingTransformMatrix(TransformMatrix, FFilterExponent,1);
  FormingWorkHMatrix(WorkHMatrix, FFilterExponent);
  FormingWorkHnullMatrix(WorkHnullMatrix, FFilterExponent);
  FormingWorkPMatrix(WorkHMatrix, WorkPMatrix);
end;

/////////////////////////////////
// Calculation transient state //
/////////////////////////////////
procedure TKalmanFiltering.CalcTransientState;
var
 xTransposeWorkHMatrix, xInverseWorkPMatrix, xTempMatrix: TMatrixDbl;
begin
  TransposeMatrix(WorkHMatrix, xTransposeWorkHMatrix);
  MultiplicationMatrix(xTransposeWorkHMatrix, WorkHMatrix, WorkPMatrix);
  InverseSymmetryMatrix(WorkPMatrix, xInverseWorkPMatrix);
  MultiplicationMatrix(xInverseWorkPMatrix, xTransposeWorkHMatrix, xTempMatrix);
  MultiplicationMatrix(xTempMatrix, TransientStateItems, PosterioriEstimation);

  if Assigned(FOnNextPosterioriEstimation) then
    FOnNextPosterioriEstimation(Self, PosterioriEstimation[0,0]);
  xTempMatrix := nil;
end;

///////////////////////////////////
//    //
///////////////////////////////////
procedure TKalmanFiltering.AddInputDataItem(Value: double);
var
  i, j, xFactorial: integer;
begin
  xFactorial := 0;

  if ItemCounter = 0 then
    InitializationFilter;
  if ItemCounter <= FFilterExponent then  //  
    begin
      TransientStateItems[ItemCounter, 0] := Value;
      if (ItemCounter = FFilterexponent) and (FMode = mdRecursive)  then
        begin
          CalcTransientState;
          CalcPrioriEstimation;
        end;
    end
  else      //  
    if FMode = mdRecursive then      //   
      begin
        PresentObservableValue := Value;
        CalcPosterioriEstimation;
        CalcPrioriEstimation;
      end
    else                            //  
      begin
        if ItemCounter <= FFilterExponent + FSizeTeachingSequence then //  
          begin
            //   H
            SetLength(WorkHMatrix, High(WorkHMatrix) + 2, FFilterExponent + 1);
            //   
            for i := High(WorkHMatrix) - 1 downto 0 do
              for j := 0 to FFilterExponent do
                WorkHMatrix[i + 1, j] := WorkHMatrix[i, j];
            //   
            for j := 0 to High(WorkHMatrix[0])  do
              if j = 0 then
                begin
                  xFactorial := 1;
                  WorkHMatrix[0, j] := 1;
                end
              else
                begin
                  xFactorial := xFactorial * j ;
                  WorkHMatrix[0, j] := IntPower(- ItemCounter, j) / xFactorial
              end;
            //  WorkPMatrix
            FormingWorkPMatrix(WorkHMatrix, WorkPMatrix);
            //   TransientStateItems
            SetLength(TransientStateItems, ItemCounter + 1, 1);
            TransientStateItems[ItemCounter, 0] := Value;
          end
        else       //   
          begin
            for i := 0 to FFilterExponent + FSizeTeachingSequence - 1 do
              TransientStateItems[i, 0] := TransientStateItems[i + 1, 0];
            TransientStateItems[FFilterExponent + FSizeTeachingSequence, 0] := Value;
            CalcTransientState;
            CalcPrioriEstimation;
          end;

      end;
  if Assigned(FOnNextItem) then
                FOnNextItem(Self, Value);
  Inc(ItemCounter);
end;

/////////////////////////////////////
// To procced along the polynomial //
/////////////////////////////////////
procedure TKalmanFiltering.GoAlongPolynomial;
begin
  CopyMatrix(PrioriEstimation, PosterioriEstimation);
  CalcPrioriEstimation;
end;

////////////////////////////////
// Get priori estimation item //
////////////////////////////////
function TKalmanFiltering.GetPrioriEstimationItem(Index: integer): double;
begin
  if ItemCounter >= FFilterExponent then
    if Index <= FFilterExponent then
      Result := PrioriEstimation[Index,0]
    else
      Result := 0
  else
    Result := 0;
end;

////////////////////////////////////
// Get posteriori estimation item //
////////////////////////////////////
function TKalmanFiltering.GetPosterioriEstimationItem(Index: integer): double;
begin
  if ItemCounter >= FFilterExponent then
    if Index <= FFilterExponent then
      Result := PosterioriEstimation[Index,0]
    else
      Result := 0
  else
    Result := 0;
end;

/////////////////////////////////
// Calculate priori estimation //
/////////////////////////////////
function TKalmanFiltering.CalcPrioriEstimation: Integer;
begin
  MultiplicationMatrix(TransformMatrix, PosterioriEstimation, PrioriEstimation);
  if Assigned(FOnNextPrioriEstimation) then
    FOnNextPrioriEstimation(Self, PrioriEstimation[0,0]);
  Result := 0;
end;

/////////////////////////////////////
//    //
/////////////////////////////////////
function TKalmanFiltering.CalcPosterioriEstimation: Integer;
var
  i, j: Integer;
  xTransposeWorkHnullMatrix: TMatrixDbl;
  xTempMatrix1, xTempMatrix2: TMatrixDbl;
begin

  TransposeMatrix(WorkHnullMatrix, xTransposeWorkHnullMatrix);
  if ItemCounter < (FSizeTeachingSequence + FFilterExponent) then
    begin
      FormingWorkNMatrix(WorkNMatrix, ItemCounter);
      TransposeMatrix(WorkNMatrix, TransposeWorkNMatrix);
      MultiplicationMatrix(TransposeWorkNMatrix, WorkNMatrix, xTempMatrix1);
      SumMatrix(WorkPMatrix, xTempMatrix1, xTempMatrix2);
      CopyMatrix(xTempMatrix2, WorkPMatrix);
      InverseSymmetryMatrix(WorkPMatrix, InverseWorkPMatrix);
    end;

  //InverseSymmetryMatrix(WorkPMatrix, xTempMatrix1);
  //MultiplicationMatrix(xTempMatrix1, xTransposeWorkHnullMatrix,xTempMatrix2);
  MultiplicationMatrix(InverseWorkPMatrix, xTransposeWorkHnullMatrix,xTempMatrix2);

  for i :=0  to High(xTempMatrix2) do
    for j := 0 to High(xTempMatrix2[0]) do
      xTempMatrix2[i, j] := xTempMatrix2[i, j] * (PresentObservableValue - PrioriEstimation[0,0]);

  SumMatrix(PrioriEstimation, xTempMatrix2, PosterioriEstimation);

  if Assigned(FOnNextPosterioriEstimation) then
    FOnNextPosterioriEstimation(Self, PosterioriEstimation[0,0]);

  Result := 0;
end;
//////////////////////////////
// Forming transform matrix //
//////////////////////////////
function TKalmanFiltering.FormingTransformMatrix(var TransformMatrix: TMatrixDbl; FilterExponent: Integer; TimeInterval: Integer): Integer;
var
  i, j: Integer;
  xFactorial: Integer;
begin
  xFactorial := 0;
  TransformMatrix := nil;
  SetLength(TransformMatrix, FilterExponent + 1, FilterExponent +1);
  //  
  for i := 0 to High(TransformMatrix) do
    for j := 0 to High(TransformMatrix) do
      TransformMatrix[i, j] := 0;
  //   
  for i := 0 to High(TransformMatrix) do
    for j := i to High(TransformMatrix) do
      if j = i then
        begin
          xFactorial := 1;
          TransformMatrix[i, j] := 1;
        end
      else
        begin
          xFactorial := xFactorial * (j - i);
          TransformMatrix[i, j] := IntPower(TimeInterval , j - i) / xFactorial
        end;
  Result := 0;
end;

///////////////////////////
// Forming Work Matrix H //
///////////////////////////
function TKalmanFiltering.FormingWorkHMatrix(var WorkHMatrix: TMatrixDbl; FilterExponent: Integer): Integer;
var
  i, j: Integer;
  xFactorial, xTimeInterval: Integer;
begin
  xFactorial := 0;
  WorkHMatrix := nil;
  SetLength(WorkHMatrix, FilterExponent + 1, FilterExponent +1);
  //  H 
  xTimeInterval := FilterExponent;
  for i := 0 to High(WorkHMatrix) do
    begin
      for j := 0 to High(WorkHMatrix) {- i} do
        if j = 0 then
          begin
            xFactorial := 1;
            WorkHMatrix[i, j] := 1;
          end
        else
          begin
            xFactorial := xFactorial * j ;
            WorkHMatrix[i, j] := IntPower(- xTimeInterval, j) / xFactorial
          end;
      Dec(xTimeInterval);
    end;

  Result := 0;
end;

//////////////////////////////
// Forming Work Matrix N    //
//////////////////////////////
function TKalmanFiltering.FormingWorkNMatrix(var WorkNMatrix: TMatrixDbl; ValueN: integer): integer;
var
  i, xFactorial: integer;
begin
  WorkNMatrix := nil;
  SetLength(WorkNMatrix, 1, FFilterExponent + 1);
  xFactorial := 1;
  for i := 0 to FFilterExponent do
    begin
      if i > 1 then
        xFactorial := xFactorial * i ;
      WorkNMatrix[0,i] := IntPower(-ValueN, i) / xFactorial;
    end;
  Result := 0;
end;

////////////////////////////
// Forming Work Matriz P  //
////////////////////////////
function TKalmanFiltering.FormingWorkPMatrix(WorkHMatrix: TMatrixDbl; var WorkPMatrix: TMatrixDbl): Integer;
var
  xTransposeWorkHMatrix: TMatrixDbl;
begin
  TransposeMatrix(WorkHMatrix, xTransposeWorkHMatrix);
  MultiplicationMatrix(xTransposeWorkHMatrix, WorkHMatrix, WorkPMatrix);
  InverseSymmetryMatrix(WorkPMatrix, InverseWorkPMatrix);
  Result := 0;
end;

///////////////////////////////
// Forming Work Matrix Hnull //
///////////////////////////////
function TKalmanFiltering.FormingWorkHnullMatrix(var WorkHnullMatrix: TMatrixDbl; FilterExponent: Integer): Integer;
var
  i: Integer;
begin
  WorkHnullMatrix := nil;
  SetLength(WorkHnullMatrix, 1, FilterExponent + 1);
  for i := 0 to High(WorkHnullMatrix) do
    if i = 0 then
      WorkHnullMatrix[0, i] := 1
    else
      WorkHnullMatrix[0, i] := 0;
  Result := 0;
end;


///////////////////////////////////////////////////////////////////////////
//      LU - .      //
//   - InputMatrix,  - OutputMatrix                 //
///////////////////////////////////////////////////////////////////////////
function TKalmanFiltering.InverseSymmetryMatrix(InputMatrix: TMatrixDbl; var OutputMatrix: TMatrixDbl): Integer;
var
  i, j, k: integer;
  xNRows: integer;
  UMatrix, LMatrix: TMatrixDbl;
  YVector: array of Extended;
  xWorkMatrix: TMatrixDbl;
  xTemp: Extended;
begin
  //  
  xNRows := High(InputMatrix);
  SetLength(UMatrix, xNRows + 1, xNRows + 1);
  SetLength(LMatrix, xNRows + 1, xNRows + 1);
  SetLength(xWorkMatrix, xNRows + 1, xNRows + 1);
  SetLength(YVector, xNRows + 1);
  CopyMatrix(InputMatrix, xWorkMatrix);


  // UP - 
  for k := 0 to xNRows do
    begin
      UMatrix[k, k] := xWorkMatrix[k, k];
      for i := k + 1 to xNRows do
        begin
          LMatrix[i, k] := xWorkMatrix[i, k] / UMatrix[k, k]; // Lik << Vi
          UMatrix[k, i] := xWorkMatrix[k, i];                 // Uki << Wi
        end;
      for i := k + 1 to xNRows do
        for j := k + 1 to xNRows do
          xWorkMatrix[i, j] := xWorkMatrix[i, j] - LMatrix[i, k] * UMatrix[k,j];
    end;
  //   
  for i := 0 to xNRows do
    LMatrix[i,i] := 1;

  // LUP - Solve  -   
  OutputMatrix := Nil;
  SetLength(OutputMatrix, xNRows + 1, xNRows + 1);
  for i := 0 to xNRows do
    begin
      YVector[i] := 0;
      OutputMatrix[i, i] := 1;
    end;

  for k := 0 to xNRows do
    begin
      for i := 0 to xNRows do
        YVector[i] := 0;

      //  
      for i := 0 to xNRows do
        begin
          xTemp := 0;
          for j := 0 to i - 1 do
            xTemp := xTemp + LMatrix[i, j] * YVector[j];
          YVector[i] := OutputMatrix[i, k] - xTemp;
        end;

      //  
      for i := xNRows downto 0 do
        begin
          xTemp := 0;
          for j := i + 1 to xNRows do
            xTemp := xTemp + UMatrix[i, j] * OutputMatrix[j, k] ;
          OutputMatrix[i, k] := (YVector[i] - xTemp) / UMatrix[i,i];
        end;
    end;
  // end LUP - Solve

  Result := 0;
end;

//////////////////////////////////////////////////
//     //
//////////////////////////////////////////////////
function TKalmanFiltering.TransposeMatrix(InputMatrix: TMatrixDbl; var OutputMatrix : TMatrixDbl): Integer;
var
  i, j : Integer;
begin
  OutputMatrix := nil;
  SetLength(OutputMatrix, High(InputMatrix[0]) + 1, High(InputMatrix) + 1);
  For i := 0 to High(InputMatrix[0]) do // i -    
    For j := 0 to High(InputMatrix) do // j -    
      OutputMatrix[i, j] := InputMatrix[j, i];
  Result := 0;
end;

//////////////////////
//   //
//////////////////////
function TKalmanFiltering.MultiplicationMatrix(MultiplicandMatrix, MultiplierMatrix: TMatrixDbl; var ProductMatrix: TMatrixDbl): Integer;
var
  i, j, n : Integer;

begin
  ProductMatrix := nil;
  //           -  
  if High(MultiplicandMatrix[0]) = High(MultiplierMatrix) then
    begin
      SetLength(ProductMatrix, High(MultiplicandMatrix) + 1, High(MultiplierMatrix[0]) +1);
      for i := 0 to High(MultiplicandMatrix)do
        for j := 0 to High(MultiplierMatrix[0]) do
          begin
            ProductMatrix[i, j] := 0;
            for n := 0 to High(MultiplierMatrix) do
              ProductMatrix[i, j] := ProductMatrix[i, j] + MultiplicandMatrix[i, n] * MultiplierMatrix[n, j];
          end;
      Result := 0;
    end
  else // 
    Result := -1;
end;

//////////////////
// Sum matrix's //
//////////////////
function TKalmanFiltering.SumMatrix(Item1Matrix: TMatrixDbl; Item2Matrix: TMatrixDbl; var SumMatrix: TMatrixDbl): Integer;
var
  i, j: Integer;
begin
  SumMatrix := nil;
  if (High(Item1Matrix) = High(Item2Matrix)) and (High(Item1Matrix[0]) = High(Item2Matrix[0])) then
    begin
      SetLength(SumMatrix, High(Item1Matrix) + 1, High(Item1Matrix[0]) + 1);
      for i := 0 to High(Item1Matrix) do
        for j := 0 to High(Item1Matrix[0]) do
          SumMatrix[i,j] := Item1Matrix[i,j] + Item2Matrix[i, j];
      Result := 0;
    end
  else
    Result := -1;
end;

/////////////////
// Copy Matrix //
/////////////////
function TKalmanFiltering.CopyMatrix(SourceMatrix: TMatrixDbl; var DestinationMatrix: TMatrixDbl): Integer;
var
  i, j: Integer;
begin
  DestinationMatrix := Nil;
  SetLength(DestinationMatrix, High(SourceMatrix) + 1, High(SourceMatrix[0]) +1);
  for i := 0 to High(SourceMatrix) do
    for j := 0 to High(SourceMatrix[0]) do
      DestinationMatrix[i,j] := SourceMatrix[i,j];
  Result := 0;
end;


end.
